==> nrgten-0.1-py3-none-any.whl: nrgten/metrics.py <==
from nrgten.encom import ENCoM
from nrgten.anm import ANM
import numpy as np
import sys
import glob
import os
from scipy.stats import pearsonr
from multiprocessing import Pool, current_process


def get_transit_probs(enm_a, enm_b, gamma=1, alignment=None):
    """Computes the transition 'probabilities' of reaching state B from state A and vice versa.

    Args:
        enm_a (ENM): The ENM object representing state A.
        enm_b (ENM): The ENM object representing state B.
        gamma (float, optional): The Boltzmann scaling factor.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]

    Returns:
        tuple: tuple containing:
            p_a(float): the probability of reaching state A from state B

            p_b(float): the probability of reaching state B from state A
    """
    vals_a = enm_a.eigvals[6:]
    vals_b = enm_b.eigvals[6:]
    if alignment is None:
        assert len(vals_a) == len(vals_b)
    else:
        assert len(alignment[0]) == len(alignment[1])
        assert len(vals_a) >= len(alignment[0])*3-6 and len(vals_b) >= len(alignment[0])*3-6
    pjs_a = np.exp(-1 * np.array(vals_a) / gamma)
    pjs_b = np.exp(-1 * np.array(vals_b) / gamma)
    pjs_a /= np.sum(pjs_a)
    pjs_b /= np.sum(pjs_b)
    if alignment is None:
        overlaps_a = get_overlaps(enm_b, enm_a, len(vals_a))
        overlaps_b = get_overlaps(enm_a, enm_b, len(vals_a))
    else:
        overlaps_a = get_overlaps(enm_b, enm_a, len(alignment[0]) * 3 - 6, [alignment[1], alignment[0]])
        overlaps_b = get_overlaps(enm_a, enm_b, len(alignment[0]) * 3 - 6, alignment)
    n = len(overlaps_a)
    p_a = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_a, pjs_a[:n])]))
    p_b = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_b, pjs_b[:n])]))
    return p_a, p_b


def fit(reference, target, alignment=None, filter=None):
    """Fits the target to the reference.

    Only the masses that have centers in the filter set are kept (all are kept if filter is None). An alignment can also
    be applied to select corresponding masses. It is applied after filtering.

    Note:
        Only the conformational change vector is returned, and the target ENM object is left untouched. This vector
        does not include the 6 trivial 3-D degrees of freedom (rotational and translational).

    Warning:
        The target object is manipulated during this operation. Could lead to nasty effects if used naively during
        parallel computing.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector representing the conformational change from reference to target
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter) # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    target._rotate(r)
    target._translate_xyz(transvec)

    # new target xyz after translation and rotation
    targ_xyz_before = target.get_filtered_xyz(filter)
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    conf_change = np.array(listify(np.subtract(targ_xyz, ref_xyz)))
    target._translate_xyz(-transvec)
    target._rotate(np.transpose(r))
    return conf_change


def get_amplitudes_for_fit(reference, target, n_modes, alignment=None, filter=None):
    """This gives the amplitudes to apply to each eigenvector of the target to get as close as possible to the target.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which coefficients will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector of amplitudes of length `n_modes`
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = reference.get_filtered_eigvecs_mat(alignment[0], filter)
    u, s, v = np.linalg.svd(eigvecs)
    x = np.dot(np.dot(np.transpose(v), np.diag(s)), np.dot(np.transpose(u), conf_change))
    return x[6:n_modes+6]


def cumulative_overlap(reference, target, n_modes, alignment=None, filter=None):
    """Cumulative overlap, as defined in Zimmerman and Jernigan 2014, Eq. 4.

    This computes the cumulative overlap between the `n_modes` first normal modes of the reference and the
    conformational change going from reference to target. See Eq. 4 from `https://doi.org/﻿10.1261/rna.041269.113`

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the cumulative overlap will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        float: the cumulative overlap, using the `n_modes` first normal modes
    """
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter))
    assert len(conf_change) == len(eigvecs[0])
    return _cumulative_overlap_helper(conf_change, eigvecs[6:6+n_modes])


def get_overlaps(reference, target, n_modes, alignment=None, filter=None):
    """Computes the overlaps between the `n_modes` from target and the conformational change.

    The conformational change is the vector of displacement going from `reference` to `target`, without
    translational/rotational degrees of freedom.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the overlaps will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        list: The list of overlaps, of length `n_modes`.
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter))
    assert len(conf_change) == len(eigvecs[0])
    overlaps = []
    n = n_modes + 6
    if n > len(eigvecs):
        n = len(eigvecs)
    for i in range(6, n):
        vec = np.array(eigvecs[i])
        overlaps.append(np.abs(np.dot(vec, conf_change))/(np.linalg.norm(vec)*np.linalg.norm(conf_change)))
    return overlaps


def rmsd_alignment(reference, target, alignment, filter):
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter)  # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    return rmsd


def _pca_ensemble(enm, macromols_list=None, variance_to_explain=0.99, filter=None):
    """ Principal component analysis on an ensemble of structures. Returns the necessary PCAs to explain the target
        proportion of variance and their respective proportions of variance explained as:
        (explained variances, PCAs) (1 PCA per row).
    """
    if macromols_list is None:
        macromols_list = enm.mols[1:]
    refmol = enm.mol
    if filter is None:
        coords = np.zeros((3*len(refmol.masses), len(macromols_list)+1))
    else:
        coords = np.zeros((3*refmol.get_filtered_n(filter), len(macromols_list) + 1))
    for i, mol in enumerate([refmol] + macromols_list):
        transvec, r = None, None
        if i > 0:
            if not mol.solved:
                mol.solve()
            rmsd, r, transvec = _fit_to_reference_macromol_DEPRECATED(refmol, mol)
        if filter is None:
            coords[:,i] = mol.get_3n_vector()
        else:
            coords[:, i] = mol.get_filtered_3n_vector(filter)
        if i > 0:
            mol.translate_xyz(-transvec)
            mol.rotate(np.transpose(r))
    # Centering the data
    means = np.zeros((len(coords)))
    for i in range(len(coords)):
        means[i] = np.mean(coords[i])
        coords[i] -= means[i]
    covmat = np.cov(coords)
    eigvals, eigvecsmat = np.linalg.eigh(covmat)
    eigvecs = np.transpose(eigvecsmat)
    s = np.sum(eigvals)
    proportion_variance = eigvals/s
    total_var = 0
    n_significant = None
    for i in range(len(proportion_variance)-1, -1, -1):
        total_var += proportion_variance[i]
        if total_var > variance_to_explain:
            n_significant = len(proportion_variance) - i
            break
    return np.flip(proportion_variance, axis=0)[0:n_significant], np.flip(eigvecs, axis=0)[0:n_significant]

def _fit_to_reference_macromol_DEPRECATED(ref, target):
    rmsd, r, transvec = rmsd_kabsch(ref.masscoords, target.masscoords)
    target.rotate(r)
    target.translate_xyz(transvec)
    return rmsd, r, transvec


def _cumulative_overlap_helper(ref_vector, vectors_list):
    s = 0
    ref_norm = np.linalg.norm(ref_vector)
    for vec in vectors_list:
        s += np.square(np.absolute(np.dot(ref_vector, vec))/(np.linalg.norm(vec)*ref_norm))
    return np.sqrt(s)


def _rmsip(ref_vectors, vectors_list):
    """ Root mean square inner product, as defined in Leo-Macis et al. 2005.
    """
    s = 0
    for r in ref_vectors:
        for v in vectors_list:
            s += np.square(np.dot(r, v))
    return np.sqrt(s/len(ref_vectors))


def _write_model(enm, count, fh):
    """ Writes one model to the given filehandle, with model number count.
    """
    fh.write("MODEL     {:>4}\n".format(count))
    enm.write_to_filehandle(fh)
    fh.write("ENDMDL\n")


def _motion(enm, mode_index, stepsize, maxrmsd, filename):
    """ Writes a PDB file displaying the motion of a normal mode, using MODEL records.
        Important note: mode_index starts at 1, which is the first nontrivial motion (7th normal mode).
    """
    mode = np.copy(enm.eigvecs[mode_index+5])
    modermsd = _rmsd_of_3n_vector(mode)
    mode *= stepsize/modermsd
    nsteps = int(maxrmsd/stepsize)
    count = 1
    offset = 0
    with open(filename, "w") as fh:
        _write_model(enm, count, fh)
        for i, sign in enumerate([1, -1, 1, -1]):
            mode *= sign
            if i == 3:
                offset = 1
            for i in range(nsteps - offset):
                enm._translate_3n_vector(mode)
                _write_model(enm, count, fh)
                count += 1


def _rmsd_of_3n_vector(vec):
    dists = np.zeros((int(len(vec)/3)))
    for i in range(0, len(vec), 3):
        dists[int(i/3)] = vec[i]**2 + vec[i+1]**2 + vec[i+2]**2
    return np.sqrt(np.mean(dists))


def listify(matNx3):
    """ Takes an N x 3 matrix and makes it into a list (eigenvector-like).
    """
    l = []
    for row in matNx3:
        assert len(row) == 3
        for e in row:
            l.append(e)
    return l


def rmsd_kabsch(q, p):
    """ Kabsch algorithm for RMSD. Returns RMSD, rotation matrix and translation
        vector to optimally superpose p into q. See https://en.wikipedia.org/wiki/Kabsch_algorithm
    """
    assert len(p) == len(q)  # make sure that p and q are Nx3 matrices (x,y,z coordinates of N points)
    for i in range(len(p)):
        assert len(p[i]) == 3
        assert len(q[i]) == 3
    p = clone_nx3(p)
    q = clone_nx3(q)
    p_centroid = get_centroid(p)
    q_centroid = get_centroid(q)
    p_centered = subtract_centroid(p, p_centroid)
    q_centered = subtract_centroid(q, q_centroid)
    h = get_cross_covariance_matrix(p_centered, q_centered)
    u, s, vt = np.linalg.svd(h)
    v = np.transpose(vt)
    ut = np.transpose(u)
    d = np.linalg.det(np.dot(v, ut))
    x = np.zeros((3, 3))
    x[0][0] = 1
    x[1][1] = 1
    x[2][2] = d
    r = np.dot(np.dot(v, x), ut)
    rotated_centered_p = np.dot(p_centered, np.transpose(r))
    translation_vector = q_centroid - p_centroid
    rmsd = np.sqrt(np.mean(np.square(get_distance_vector(q_centered, rotated_centered_p))))
    return rmsd, r, translation_vector


def get_distance_vector(q, p):
    assert len(p) == len(q)
    d = np.zeros((len(p)))
    for i in range(len(p)):
        d[i] = np.linalg.norm(p[i]-q[i])
    return d


def get_cross_covariance_matrix(p, q):
    return np.dot(np.transpose(p), q)


def subtract_centroid(p, centroid):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j] - centroid[j]
    return new_p


def clone_nx3(p):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j]
    return new_p


def get_centroid(p):
    """ Returns the centroid of a set of point p (in the form of an Nx3 matrix).
    """
    n = len(p)
    xyz = np.array([0, 0, 0])
    for i in range(n):
        xyz = xyz + np.array([p[i][0], p[i][1], p[i][2]])
    return xyz/n




if __name__ == "__main__":
    # test_boltzmann_squares()


    open = ENCoM("../../tests/open_clean.pdb")
    closed = ENCoM("../../tests/closed_clean.pdb")
    print(get_amplitudes_for_fit(open, closed, 10))
    # print(overlap(open, closed, 10))
    # print(overlap(closed, open, 10))
    # print(open.compute_bfactors()[:15])

==> nrgten-0.1.2-py3-none-any.whl: nrgten/metrics.py <==
from nrgten.encom import ENCoM
from nrgten.anm import ANM
import numpy as np
# import sys
# import glob
# import os
# from scipy.stats import pearsonr
# from multiprocessing import Pool, current_process


def get_transit_probs(enm_a, enm_b, gamma=1, alignment=None):
    """Computes the transition 'probabilities' of reaching state B from state A and vice versa.

    Args:
        enm_a (ENM): The ENM object representing state A.
        enm_b (ENM): The ENM object representing state B.
        gamma (float, optional): The Boltzmann scaling factor.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]

    Returns:
        tuple: tuple containing:
            p_a(float): the probability of reaching state A from state B

            p_b(float): the probability of reaching state B from state A
    """
    vals_a = enm_a.eigvals[6:]
    vals_b = enm_b.eigvals[6:]
    if alignment is None:
        assert len(vals_a) == len(vals_b)
    else:
        assert len(alignment[0]) == len(alignment[1])
        assert len(vals_a) >= len(alignment[0])*3-6 and len(vals_b) >= len(alignment[0])*3-6
    pjs_a = np.exp(-1 * np.array(vals_a) / gamma)
    pjs_b = np.exp(-1 * np.array(vals_b) / gamma)
    pjs_a /= np.sum(pjs_a)
    pjs_b /= np.sum(pjs_b)
    if alignment is None:
        overlaps_a = get_overlaps(enm_b, enm_a, len(vals_a))
        overlaps_b = get_overlaps(enm_a, enm_b, len(vals_a))
    else:
        overlaps_a = get_overlaps(enm_b, enm_a, len(alignment[0]) * 3 - 6, [alignment[1], alignment[0]])
        overlaps_b = get_overlaps(enm_a, enm_b, len(alignment[0]) * 3 - 6, alignment)
    n = len(overlaps_a)
    p_a = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_a, pjs_a[:n])]))
    p_b = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_b, pjs_b[:n])]))
    return p_a, p_b


def fit(reference, target, alignment=None, filter=None):
    """Fits the target to the reference.

    Only the masses that have centers in the filter set are kept (all are kept if filter is None). An alignment can also
    be applied to select corresponding masses. It is applied after filtering.

    Note:
        Only the conformational change vector is returned, and the target ENM object is left untouched. This vector
        does not include the 6 trivial 3-D degrees of freedom (rotational and translational).

    Warning:
        The target object is manipulated during this operation. Could lead to nasty effects if used naively during
        parallel computing.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector representing the conformational change from reference to target
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter) # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    target._rotate(r)
    target._translate_xyz(transvec)

    # new target xyz after translation and rotation
    targ_xyz_before = target.get_filtered_xyz(filter)
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    conf_change = np.array(listify(np.subtract(targ_xyz, ref_xyz)))
    target._translate_xyz(-transvec)
    target._rotate(np.transpose(r))
    return conf_change


def get_amplitudes_for_fit(reference, target, n_modes, alignment=None, filter=None):
    """This gives the amplitudes to apply to each eigenvector of the target to get as close as possible to the target.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which coefficients will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector of amplitudes of length `n_modes`
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = reference.get_filtered_eigvecs_mat(alignment[0], filter)
    u, s, v = np.linalg.svd(eigvecs)
    x = np.dot(np.dot(np.transpose(v), np.diag(s)), np.dot(np.transpose(u), conf_change))
    return x[6:n_modes+6]


def cumulative_overlap(reference, target, n_modes, alignment=None, filter=None):
    """Cumulative overlap, as defined in Zimmerman and Jernigan 2014, Eq. 4.

    This computes the cumulative overlap between the `n_modes` first normal modes of the reference and the
    conformational change going from reference to target. See Eq. 4 from `https://doi.org/﻿10.1261/rna.041269.113`

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the cumulative overlap will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        float: the cumulative overlap, using the `n_modes` first normal modes
    """
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter))
    assert len(conf_change) == len(eigvecs[0])
    return _cumulative_overlap_helper(conf_change, eigvecs[6:6+n_modes])


def get_overlaps(reference, target, n_modes, alignment=None, filter=None):
    """Computes the overlaps between the `n_modes` from target and the conformational change.

    The conformational change is the vector of displacement going from `reference` to `target`, without
    translational/rotational degrees of freedom.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the overlaps will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        list: The list of overlaps, of length `n_modes`.
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter))
    assert len(conf_change) == len(eigvecs[0])
    overlaps = []
    n = n_modes + 6
    if n > len(eigvecs):
        n = len(eigvecs)
    for i in range(6, n):
        vec = np.array(eigvecs[i])
        overlaps.append(np.abs(np.dot(vec, conf_change))/(np.linalg.norm(vec)*np.linalg.norm(conf_change)))
    return overlaps


def rmsd_alignment(reference, target, alignment, filter):
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter)  # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    return rmsd


def _pca_ensemble(enm, macromols_list=None, variance_to_explain=0.99, filter=None):
    """ Principal component analysis on an ensemble of structures. Returns the necessary PCAs to explain the target
        proportion of variance and their respective proportions of variance explained as:
        (explained variances, PCAs) (1 PCA per row).
    """
    if macromols_list is None:
        macromols_list = enm.mols[1:]
    refmol = enm.mol
    if filter is None:
        coords = np.zeros((3*len(refmol.masses), len(macromols_list)+1))
    else:
        coords = np.zeros((3*refmol.get_filtered_n(filter), len(macromols_list) + 1))
    for i, mol in enumerate([refmol] + macromols_list):
        transvec, r = None, None
        if i > 0:
            if not mol.solved:
                mol.solve()
            rmsd, r, transvec = _fit_to_reference_macromol_DEPRECATED(refmol, mol)
        if filter is None:
            coords[:,i] = mol.get_3n_vector()
        else:
            coords[:, i] = mol.get_filtered_3n_vector(filter)
        if i > 0:
            mol.translate_xyz(-transvec)
            mol.rotate(np.transpose(r))
    # Centering the data
    means = np.zeros((len(coords)))
    for i in range(len(coords)):
        means[i] = np.mean(coords[i])
        coords[i] -= means[i]
    covmat = np.cov(coords)
    eigvals, eigvecsmat = np.linalg.eigh(covmat)
    eigvecs = np.transpose(eigvecsmat)
    s = np.sum(eigvals)
    proportion_variance = eigvals/s
    total_var = 0
    n_significant = None
    for i in range(len(proportion_variance)-1, -1, -1):
        total_var += proportion_variance[i]
        if total_var > variance_to_explain:
            n_significant = len(proportion_variance) - i
            break
    return np.flip(proportion_variance, axis=0)[0:n_significant], np.flip(eigvecs, axis=0)[0:n_significant]

def _fit_to_reference_macromol_DEPRECATED(ref, target):
    rmsd, r, transvec = rmsd_kabsch(ref.masscoords, target.masscoords)
    target.rotate(r)
    target.translate_xyz(transvec)
    return rmsd, r, transvec


def _cumulative_overlap_helper(ref_vector, vectors_list):
    s = 0
    ref_norm = np.linalg.norm(ref_vector)
    for vec in vectors_list:
        s += np.square(np.absolute(np.dot(ref_vector, vec))/(np.linalg.norm(vec)*ref_norm))
    return np.sqrt(s)


def _rmsip(ref_vectors, vectors_list):
    """ Root mean square inner product, as defined in Leo-Macis et al. 2005.
    """
    s = 0
    for r in ref_vectors:
        for v in vectors_list:
            s += np.square(np.dot(r, v))
    return np.sqrt(s/len(ref_vectors))


def _write_model(enm, count, fh):
    """ Writes one model to the given filehandle, with model number count.
    """
    fh.write("MODEL     {:>4}\n".format(count))
    enm.write_to_filehandle(fh)
    fh.write("ENDMDL\n")


def _motion(enm, mode_index, stepsize, maxrmsd, filename):
    """ Writes a PDB file displaying the motion of a normal mode, using MODEL records.
        Important note: mode_index starts at 1, which is the first nontrivial motion (7th normal mode).
    """
    mode = np.copy(enm.eigvecs[mode_index+5])
    modermsd = _rmsd_of_3n_vector(mode)
    mode *= stepsize/modermsd
    nsteps = int(maxrmsd/stepsize)
    count = 1
    offset = 0
    with open(filename, "w") as fh:
        _write_model(enm, count, fh)
        for i, sign in enumerate([1, -1, 1, -1]):
            mode *= sign
            if i == 3:
                offset = 1
            for i in range(nsteps - offset):
                enm._translate_3n_vector(mode)
                _write_model(enm, count, fh)
                count += 1


def _rmsd_of_3n_vector(vec):
    dists = np.zeros((int(len(vec)/3)))
    for i in range(0, len(vec), 3):
        dists[int(i/3)] = vec[i]**2 + vec[i+1]**2 + vec[i+2]**2
    return np.sqrt(np.mean(dists))


def listify(matNx3):
    """ Takes an N x 3 matrix and makes it into a list (eigenvector-like).
    """
    l = []
    for row in matNx3:
        assert len(row) == 3
        for e in row:
            l.append(e)
    return l


def rmsd_kabsch(q, p):
    """ Kabsch algorithm for RMSD. Returns RMSD, rotation matrix and translation
        vector to optimally superpose p into q. See https://en.wikipedia.org/wiki/Kabsch_algorithm
    """
    assert len(p) == len(q)  # make sure that p and q are Nx3 matrices (x,y,z coordinates of N points)
    for i in range(len(p)):
        assert len(p[i]) == 3
        assert len(q[i]) == 3
    p = clone_nx3(p)
    q = clone_nx3(q)
    p_centroid = get_centroid(p)
    q_centroid = get_centroid(q)
    p_centered = subtract_centroid(p, p_centroid)
    q_centered = subtract_centroid(q, q_centroid)
    h = get_cross_covariance_matrix(p_centered, q_centered)
    u, s, vt = np.linalg.svd(h)
    v = np.transpose(vt)
    ut = np.transpose(u)
    d = np.linalg.det(np.dot(v, ut))
    x = np.zeros((3, 3))
    x[0][0] = 1
    x[1][1] = 1
    x[2][2] = d
    r = np.dot(np.dot(v, x), ut)
    rotated_centered_p = np.dot(p_centered, np.transpose(r))
    translation_vector = q_centroid - p_centroid
    rmsd = np.sqrt(np.mean(np.square(get_distance_vector(q_centered, rotated_centered_p))))
    return rmsd, r, translation_vector


def get_distance_vector(q, p):
    assert len(p) == len(q)
    d = np.zeros((len(p)))
    for i in range(len(p)):
        d[i] = np.linalg.norm(p[i]-q[i])
    return d


def get_cross_covariance_matrix(p, q):
    return np.dot(np.transpose(p), q)


def subtract_centroid(p, centroid):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j] - centroid[j]
    return new_p


def clone_nx3(p):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j]
    return new_p


def get_centroid(p):
    """ Returns the centroid of a set of point p (in the form of an Nx3 matrix).
    """
    n = len(p)
    xyz = np.array([0, 0, 0])
    for i in range(n):
        xyz = xyz + np.array([p[i][0], p[i][1], p[i][2]])
    return xyz/n




if __name__ == "__main__":
    # test_boltzmann_squares()


    open = ENCoM("../../tests/open_clean.pdb")
    closed = ENCoM("../../tests/closed_clean.pdb")
    print(get_amplitudes_for_fit(open, closed, 10))
    # print(overlap(open, closed, 10))
    # print(overlap(closed, open, 10))
    # print(open.compute_bfactors()[:15])
